#include <stdio.h>
#ifdef STDC_HEADERS
#  include <stdlib.h>
#else
   char *malloc (), *realloc ();
#endif
#ifdef USG
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <ctype.h>

#include <coco.h>

char *libcoco_version = "2.0";

extern Lisp_Object code_category[IDX_BIN + 1];
extern char code_priority_category[IDX_BIN + 1];

coding_type coding_system_table[CODING_SYSTEM_COUNT];
int max_coding_system = 0;
int max_coding_system_extra;

#ifndef CODINGS_PATH
#define CODINGS_PATH ".,/usr/local/lib/mule/19.27/etc"
#endif

#ifndef CODINGS_NAME
#define CODINGS_NAME "CODINGS"
#endif

#ifndef CHARSETS_PATH
#define CHARSETS_PATH ".,/usr/local/lib/mule/19.27etc"
#endif

#ifndef CHARSETS_NAME
#define CHARSETS_NAME "CHARSETS"
#endif

int coco_error;
char coco_error_msg[256];

/***************************************************
 * General routine for reading CHARSETS and CODINGS *
 ***************************************************/

#ifndef MAXPATHLEN
#define MAXPATHLEN 256
#endif

#define LINE_SIZE 4096

FILE *
open_file (dirs, file)
     char *dirs, *file;
{
  char path[MAXPATHLEN];	/* 93.2.10 by I.Hirakura */
#ifdef MSDOS
  char drive = 0;

  if (strlen(file) >= 2 && file[1] == ':') {
    drive = tolower(file[0]);
    dirs = "", file += 2;
  }
#endif
  while (file[0] == '/') dirs = "", file++; /* file is absolute path */

  while (1)
    {
      char *tail;
      int len;
      FILE *fp;

#ifdef MSDOS
      if (strlen(dirs) < 2 || dirs[1] != ':') {
	char *temp = alloca (strlen(dirs)+3);
	
	temp[0] = drive, temp[1] = ':';
	dirs = strcpy (&temp[2], dirs);
      }
#endif
      (tail = index (dirs, ',')) || (tail = index (dirs, '\0'));

      len = tail - dirs;
      strncpy (path, dirs, len);

      while (path[--len] == '/'); /* back to non-slash */
      path[++len] = '/';
      strcpy (&path[++len], file);

/* 92.10.27 by M.Higashida */
#ifdef MSDOS
      fp = fopen (path, "rt");
#else
      fp = fopen (path, "r");
#endif
/* end of patch */
      if (fp) return fp;
 
      if (*tail == '\0') return (FILE *)0;
      dirs = tail + 1;
    }
}

#define PROCEED_CHAR(c) \
  if (!(p1 = index (p0, c))) goto invalid_entry

/********************
 * Reading CHARSETS *
 ********************/

set_charsets_param(line)
     char *line;
{
  char *p0 = line, *p1;
  int i;
  unsigned char lc, bytes, clm, type, graphic, final, direction;
  char *doc;

  lc = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  bytes = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  clm = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  type = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  graphic = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  final = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  direction = atoi(p0);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  if (p1 = index(p0, '\n')) *p1++ = NULL;
  else p1 = index(p0, '\0') + 1;
  doc = (char *)malloc(p1 - p0);
  bcopy(p0, doc, p1 - p0);

  update_mc_table(lc, bytes, clm, type, graphic, final, direction, doc);
  return 0;

 invalid_entry:
  coco_error = COCO_ERROR_INVALID_CHARSETS;
  sprintf(coco_error_msg, "Invalid entry in CHARSETS file: %s", line);
  return -1;
}
 
static CCL_PROGRAM *ccl_programs[128];

set_ccl_program_param(line)
     char *line;
{
  int lc, len, i, j;
  char *p, *p1;
  Lisp_Object ccl;

  lc = atoi(line) & 0x7F;
  p = p1 = index(line, ':') + 1;
  if (!p) return -1;
  len = 0;
  while (*p1) if (*p1++ == ' ') len++;
  ccl = make_vector (len);
  for (i = 0; i < len; i++) {
    sscanf(p, "%x", XVECTOR (ccl)->contents + i);
    p = index (p + 1, ' ');
  }
  ccl_programs[lc] = (CCL_PROGRAM *) malloc (sizeof (CCL_PROGRAM));
  set_ccl_program(ccl_programs[lc], ccl);
  return 0;
}

charsets_initialize(charsets_path, charsets_name)
     char *charsets_path, *charsets_name;
{
  FILE *fp;
  char line[LINE_SIZE];
  int i, status;

  if (!charsets_path) charsets_path = CHARSETS_PATH;
  if (!charsets_name) charsets_name = CHARSETS_NAME;
  fp = open_file(charsets_path, charsets_name);
  if (!fp) {
    coco_error = COCO_ERROR_NO_CHARSETS;
    sprintf(coco_error_msg, "File %s not in the path %s",
	    charsets_name, charsets_path);
    return -1;
  }
  init_charset_once();
  status = 0;
  for (i = 0; i < 128; i++)
    ccl_programs[i] = NULL;
  while (fgets(line, sizeof line, fp)) {
    if (line[0] == '#') {
      if (status > 0) status++;
      continue;
    }
    line[strlen(line) - 1] = 0;
    if (status <= 1) {
      status = 1;
      if (set_charsets_param(line) < 0) return -1;
    } else {
      if (set_ccl_program_param(line) < 0) return -1;
    }
  }
  return 0;
}

/******************
 * Reading CODINGS *
 ******************/

set_coding_system_param(line, idx)
     char *line;
     int idx;
{
  char *p0 = line, *p1;
  coding_type *cs = coding_system_table + idx;
  int i;

  PROCEED_CHAR(':');
  *p1 = NULL;
  cs->name = (char *)malloc(strlen(p0) + 1);
  strcpy(cs->name, p0);
  *p1++ = ':';

  p0 = p1;
  i = atoi(p0);
  CODE_TYPE_SET(cs, i);
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  cs->mnemonic = *p0;
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  if (isdigit(*p0)) {
    i = atoi(p0);
    CODE_FORM (cs)
      = (i==0) ? CODE_EOL_AUTO : (i==1) ? CODE_EOL_LF
	: (i==2) ? CODE_EOL_CRLF : CODE_EOL_CR;
    PROCEED_CHAR(':');
    CODE_LF (cs) = CODE_CRLF (cs) = CODE_CR (cs) = Qnil;
  } else {
    CODE_FORM (cs) = CODE_EOL_AUTO;
    PROCEED_CHAR(',');
    *p1 = NULL;
    CODE_LF (cs) = make_symbol(p0);
    *p1++ = ',';
    p0 = p1;
    PROCEED_CHAR(',');
    *p1 = NULL;
    CODE_CRLF (cs) = make_symbol(p0);
    *p1++ = ',';
    p0 = p1;
    PROCEED_CHAR(':');
    *p1 = NULL;
    CODE_CR (cs) = make_symbol(p0);
    *p1 = ':';
  }

  p0 = p1 + 1;
  if (CODE_TYPE (cs) == ISO2022)  {
    int flags[12];
    for (i = 0; i < 11; i++) {
      if (!(flags[i] = atoi(p0))) flags[i] = Qnil;
      PROCEED_CHAR(',');
      p0 = p1 + 1;
    }
    flags[i] = atoi(p0);
    CODE_LC_SET(cs, flags[0], flags[1], flags[2], flags[3]);
    CODE_FORM_SET(cs, (Lisp_Object)flags[4], (Lisp_Object)flags[5],
		  (Lisp_Object)flags[6], (Lisp_Object)flags[7],
		  (Lisp_Object)flags[8], (Lisp_Object)flags[9],
		  (Lisp_Object)flags[10], (Lisp_Object)flags[11]);
  } else if (CODE_TYPE (cs) == BIG5) {
    CODE_FORM_SET(cs, Qnil, Qnil, Qnil, Qnil, Qnil, Qnil, Qnil, Qnil);
  } else if (CODE_TYPE (cs) == CCL) {
    int len;
    char *start = p0, c;
    Lisp_Object ccl;

    len = 0;
    while ((c = *p0++) != ',') if (c == ' ') len++;
    ccl = make_vector (len);
    for (i = 0; i < len; i++) {
      sscanf(start, "%x", XVECTOR (ccl)->contents + i);
      start = index (start + 1, ' ');
    }
    set_ccl_program (&CODE_CCL_ENCODE (cs), ccl);

    start = p0;
    len = 0;
    while ((c = *p0++) != ':') if (c == ' ') len++;
    ccl = make_vector (len);
    for (i = 0; i < len; i++) {
      sscanf(start, "%x", XVECTOR (ccl)->contents + i);
      start = index (start + 1, ' ');
    }
    set_ccl_program (&CODE_CCL_DECODE (cs), ccl);
    p0 -= 2;
  }
  PROCEED_CHAR(':');

  p0 = p1 + 1;
  if (p1 = index(p0, '\n')) *p1++ = NULL;
  else p1 = index(p0, '\0') + 1;
  cs->doc = (char *)malloc(p1 - p0);
  bcopy(p0, cs->doc, p1 - p0);
  return 0;

 invalid_entry:
  coco_error = COCO_ERROR_INVALID_CODING;
  sprintf(coco_error_msg, "Invalid entry in CODINGS file: %s", line);
  return -1;
}

char *coding_category_names[IDX_BIN + 1] = {
  "*coding-category-internal*",
  "*coding-category-sjis*",
  "*coding-category-iso-7*",
  "*coding-category-iso-8-1*",
  "*coding-category-iso-8-2*",
  "*coding-category-iso-else*",
  "*coding-category-big5*",
  "*coding-category-bin*"};

set_coding_category(line, priority)
     char *line;
     int priority;
{
  char *p0 = line, *p1;
  int i;

  PROCEED_CHAR(':');
  *p1 = NULL;
  for (i = 0; i <= IDX_BIN; i++) {
    if (!strcmp(p0, XSYMBOL (code_category[i])->name))
      break;
  }
  if (i <= IDX_BIN) {
    code_priority_category[priority] = i;
    p0 = ++p1;
    if (p1 = index(p0, '\n')) *p1 = NULL;
    XSYMBOL (code_category[i])->value = make_symbol(p0);
    return 0;
  } else {
    *p1 = ':';
  invalid_entry:
    coco_error = COCO_ERROR_INVALID_CODING;
    sprintf(coco_error_msg, "Invalid entry in CODINGS file: %s", line);
    return -1;
  }
} 

codings_initialize(codings_path, codings_name)
     char *codings_path, *codings_name;
{
  FILE *fp;
  char line[LINE_SIZE];
  char *coding_category_header = "## LIST OF CODING CATEGORIES";
  int i;

  if (!codings_path) codings_path = CODINGS_PATH;
  if (!codings_name) codings_name = CODINGS_NAME;
  fp = open_file(codings_path, codings_name);
  if (!fp) {
    coco_error = COCO_ERROR_NO_CODING;
    sprintf(coco_error_msg, "File %s not in the path %s!",
	    codings_name, codings_path, 0);
    return -1;
  }
  max_coding_system = 0;
  while (fgets(line, sizeof line, fp)
	 && strncmp(line, coding_category_header,
		    strlen(coding_category_header))) {
    if (line[0] != '#') {
      if (set_coding_system_param(line, max_coding_system) < 0)
	return -1;
      max_coding_system++;
    }
  }
  {
    coding_type *cs = coding_system_table + max_coding_system;

    for (i = 0; i < max_coding_system; i++) {
      if (CODE_TYPE (&coding_system_table[i]) != CCL) {
	sprintf(line, "%sunix*", coding_system_table[i].name);
	coding_system_table[i].eol_lf = make_symbol(line);
	*cs = coding_system_table[i];
	cs->name = XSYMBOL (coding_system_table[i].eol_lf)->name;
	CODE_FORM (cs) |= CODE_EOL_LF;
	CODE_LF (cs) = CODE_CRLF (cs) = CODE_CR (cs) = Qnil;
	cs++;

	sprintf(line, "%sdos*", coding_system_table[i].name);
	coding_system_table[i].eol_crlf = make_symbol(line);
	*cs = coding_system_table[i];
	cs->name = XSYMBOL (coding_system_table[i].eol_crlf)->name;
	CODE_FORM (cs) |= CODE_EOL_CRLF;
	CODE_LF (cs) = CODE_CRLF (cs) = CODE_CR (cs) = Qnil;
	cs++;

	sprintf(line, "%smac*", coding_system_table[i].name);
	coding_system_table[i].eol_cr = make_symbol(line);
	*cs = coding_system_table[i];
	cs->name = XSYMBOL (coding_system_table[i].eol_cr)->name;
	CODE_FORM (cs) |= CODE_EOL_CR;
	CODE_LF (cs) = CODE_CRLF (cs) = CODE_CR (cs) = Qnil;
	cs++;
      }
    }
    max_coding_system_extra = cs - coding_system_table;
  }

  for (i = 0; i <= IDX_BIN; i++)
    code_category[i] = make_symbol(coding_category_names[i]);
  i = 0;
  while (fgets(line, sizeof line, fp)) {
    if (line[0] != '#') {
      if (set_coding_category(line, i) < 0)
	return -1;
      i++;
    }
  }
  return 0;
}

coco_process_ccl(buf, n)
     unsigned char *buf;
     int n;
{
  unsigned char *pend = buf + n;
  CCL_PROGRAM *ccl;

  while (buf < pend) {
    if (*buf < 0x80 || !LC_P (*buf))
      buf++;
    else {
      if (*buf < LCPRV11 && (ccl = ccl_programs[*buf & 0x7f])) {
	if (char_bytes[*buf] == TWOBYTE)
	  ccl->reg[0] = 0, ccl->reg[1] = *(buf + 1);
	else
	  ccl->reg[0] = *(buf + 1), ccl->reg[1] = *(buf + 2);
	ccl_driver(ccl, (char *)0, (char *)0, 0, 0);
	if (char_bytes[*buf] == TWOBYTE)
	  *(buf + 1) = ccl->reg[1];
	else
	  *(buf + 1) = ccl->reg[0], *(buf + 2) = ccl->reg[1];
      } else if (*buf >= LCPRV11
		 && VALID_LC (*(buf + 1))
		 && (ccl = ccl_programs[*(buf + 1) & 0x7f])) {
	if (char_bytes[*buf] == THREEBYTE)
	  ccl->reg[0] = 0, ccl->reg[1] = *(buf + 2);
	else
	  ccl->reg[0] = *(buf + 2), ccl->reg[1] = *(buf + 3);
	ccl_driver(ccl, (char *)0, (char *)0, 0, 0);
	if (char_bytes[*buf] == THREEBYTE)
	  *(buf + 2) = ccl->reg[1];
	else
	  *(buf + 2) = ccl->reg[0], *(buf + 3) = ccl->reg[1];
      }
      buf += char_bytes[*buf];
    }
  }
}

coco_find_coding(str)
     char *str;
{
  int i;

  if (*str == '*') {
    for (i = 0; i < max_coding_system_extra; i++)
      if (!strcmp(str, coding_system_table[i].name)) return i;
  } else {
    unsigned int eol_type;
    switch (str[1]) {
    case '\0': eol_type = CODE_EOL_AUTO; break;
    case '.': eol_type = CODE_EOL_LF; break;
    case ':': eol_type = CODE_EOL_CRLF; break;
    case '\'': eol_type = CODE_EOL_CR; break;
    default: return -1;
    }
    for (i = 0; i < max_coding_system; i++)
      if (coding_system_table[i].mnemonic == *str
	  && (CODE_FORM (&coding_system_table[i]) & CODE_EOL_MASK) == eol_type)
	return i;
  }
  return -1;
}

encode_code(code, mccode)
     Lisp_Object code;
     coding_type *mccode;
{
  int i;

  if ((i = coco_find_coding(XSYMBOL (code)->name)) >= 0)
    *mccode = coding_system_table[i];
  else
    CODE_TYPE_SET (mccode, NOCONV);
}

/***********************
 * Interface functions *
 ***********************/

coco_initialize()
{
  if (codings_initialize(0, 0) < 0) return -1;
  return charsets_initialize(0, 0);
}

coco_reset(inname, incode, outname, outcode)
     char *inname, *outname;
     coding_type *incode, *outcode;
{
  int i;

  if (max_coding_system == 0) {
    if (coco_initialize() < 0) return -1;
  }

  if (inname == NULL) inname = "*autoconv*";
  if ((i = coco_find_coding(inname)) < 0) {
    coco_error = COCO_ERROR_UNKNOWN_CODE;
    sprintf(coco_error_msg, "Unknown code: %s", inname);
    return -1;
  }
  *incode = coding_system_table[i];

  if (outname == NULL) outname = "*internal*";
  if ((i = coco_find_coding(outname)) < 0) {
    coco_error = COCO_ERROR_UNKNOWN_CODE;
    sprintf(coco_error_msg, "Unknown code: %s", outname);
    return -1;
  }
  *outcode = coding_system_table[i];

  return 0;
}

coco_perror()
{
  fprintf(stderr, "COCO: %s\n", coco_error_msg);
}

coco(incode, inbuf, insize, outcode, outbuf, outsize, ccl)
     coding_type *incode, *outcode;
     char *inbuf, *outbuf;
     int insize, outsize, ccl;
{
  Lisp_Object found;
  char *buf1, *buf2;
  int n;

  if (insize > 0) {
    buf1 = (char *)alloca(ENCODE_BUF_SIZE(insize, incode));
    n = encode(incode, inbuf, buf1, insize, &found);
    incode->name = (char *)XSYMBOL (found)->name;
  } else {
    buf1 = inbuf;
    n = 0;
    CODE_CNTL(outcode) |= CC_END;
  }

  if (CODE_TYPE (outcode) == ITNCODE) {
    if (ccl)
      coco_process_ccl(buf1, n);
    buf2 = buf1;
  } else {
    if ((CODE_FORM (outcode) & CODE_EOL_MASK) == CODE_EOL_AUTO
	&& (CODE_FORM (incode) & CODE_EOL_MASK) != CODE_EOL_AUTO) {
      unsigned int eol = CODE_FORM (incode) & CODE_EOL_MASK;

      CODE_FORM (outcode) |= eol;
      switch (eol) {
      case CODE_EOL_LF:
	outcode->name = (char *)XSYMBOL (outcode->eol_lf)->name; break;
      case CODE_EOL_CRLF:
	outcode->name = (char *)XSYMBOL (outcode->eol_crlf)->name; break;
      case CODE_EOL_CR:
	outcode->name = (char *)XSYMBOL (outcode->eol_cr)->name; break;
      }
    }
    buf2 = (char *)alloca(DECODE_BUF_SIZE(n, outcode));
    n = decode(outcode, buf1, buf2, n);
  }

  if (n <= outsize) {
    bcopy(buf2, outbuf, n);
  } else {
    coco_error = COCO_ERROR_OVERFLOW;
    sprintf(coco_error_msg, "Output buffer overflow");
  }

  return (n <= outsize ? n : -1);
}
